﻿/// バイト列の表現形式
/// 
/// ITweDataFormat で出力されるバイト列に対して、ヘッダフッタ等を付加したコマンド形態への変形を行う
/// 
/// 例えば以下のようなデータ Format, Payload で示されるされるデータがあった場合、
/// 78 01 00 11 22 33 44 55
/// ^^^^1 ^^^^^^^^^^^^^^^^2
/// Format            Payload
/// このバイト列に対しての表現形式を定義する。
/// 
/// 例えば clsDataPresentationAscii 形式では、以下のような文字列に表現される。
/// ":7801001122334455??[CR][LF]"
/// 
/// ?? はチェックサム
/// [CR][LF] は 0D 0A の改行コード
using System.Diagnostics;
using System.Linq;
using System.Collections.Generic;
using System.Collections;
using System;
using System.Xml.Linq;

namespace MONOW_PAL_PARSER
{
    public class ClsDataPresentationAscii : ITweDataPresentation
    {
        private ITweDataFormat _dataFormat;
        private byte[] _abyInput = new byte[1025];
        private byte _bPrev;
        private int _iLen;
        private int _iLrc;
        private E_PARSE_STATE _eState;
        private bool _bStored;

        private DateTime _timeGot;

        /// <summary>
        /// limited '0-9A-F'
        /// </summary>
        /// <param name="b"></param>
        /// <returns></returns>
        private int ConvAscToHex(byte b)
        {
            if (b >= 0x41 && b <= 0x46) return b - 0x41 + 10;
            else if (b >= 0x30 && b <= 0x39) return b - 0x30;
            else throw null;
        }
        /// <summary>
        /// limited from 0x0 to 0xF
        /// </summary>
        /// <param name="b"></param>
        /// <returns></returns>
        private int ConvHexToAsc(byte b)
        {
            if (b >= 10) return b + 0x41 - 10;
            else if (b < 16) return b + 0x30;
            else throw null;
        }

        public E_PARSE_STATE ParseByte(byte b)
        {
            if (_eState ==E_PARSE_STATE.ERR | _eState == E_PARSE_STATE.COMPLETE)
                _eState = E_PARSE_STATE.EMPTY;

            switch (_eState)
            {
                case E_PARSE_STATE.EMPTY:
                    {
                        // 先頭の : の判定
                        if (b == (int)':')
                        {
                            _eState = E_PARSE_STATE.READPAYLOAD1;
                            _iLen = 0;
                            _iLrc = 0;
                        }

                        break;
                    }

                case E_PARSE_STATE.READPAYLOAD1:
                    {
                        // HEXの読み出しまたは終端の検知
                        if (b == 0xD)
                            _eState = E_PARSE_STATE.READFOTTER;
                        else
                            _eState = E_PARSE_STATE.READPAYLOAD2;
                        break;
                    }

                case E_PARSE_STATE.READPAYLOAD2:
                    {
                        // HEXの読み出し(下位)
                        try
                        {
                            //_abyInput[_iLen] = Convert.ToByte(Convert.ToChar(_bPrev) + Convert.ToChar(b), 16);
                            _abyInput[_iLen] = (byte)((ConvAscToHex(_bPrev)<<4) + ConvAscToHex(b));
                            _iLrc = _iLrc + _abyInput[_iLen];
                            _iLen = _iLen + 1;
                            _eState = E_PARSE_STATE.READPAYLOAD1;
                        }
                        catch
                        {
                            _eState = E_PARSE_STATE.ERR;
                        }

                        break;
                    }
            }

            if (_eState == E_PARSE_STATE.READFOTTER)
            {
                if ((_iLrc & 0xFF) == 0)
                {

                    // コピー先の配列を作成する 
                    byte[] abyDat = new byte[_iLen - 1 - 1 + 1];

                    // 指定された範囲をコピーする 
                    Array.Copy(_abyInput, 0, abyDat, 0, _iLen - 1);

                    // Format/Payload を解析
                    _dataFormat = PAL_Utils.genDataFormat(ref abyDat);

                    _timeGot = DateTime.Now;

                    _eState = E_PARSE_STATE.COMPLETE;
                }
                else
                {
                    // incorrect LRC
                    _eState = E_PARSE_STATE.ERR;
                    byte byLrcCorrect;
                    byLrcCorrect = (byte)(~(_iLrc - _abyInput[_iLen - 1]) + 1);
                }
            }

            // 直前の文字
            _bPrev = b;

            // 戻り値
            return _eState;
        }

        public byte[] toByteArray()
        {
            byte[] bout;


            if (_dataFormat != null)
            {
                byte[] b;

                b = _dataFormat.toByteArray();

                if (b == null)
                    return null;

                if (b.Length > 0)
                {
                    int i;
                    int j;
                    byte lrc;
                    byte h;
                    byte l;

                    // 配列の延長
                    bout = new byte[b.Length * 2 + 5 - 1 + 1];

                    // １文字目は :
                    bout[0] = (int)':';

                    // データ形式への変換
                    lrc = 0;
                    j = 1;
                    var loopTo = b.Length - 1;
                    for (i = 0; i <= loopTo; i++)
                    {
                        lrc = (byte)(lrc + b[i]);

                        h = (byte)((b[i] & 0xF0) >> 4);
                        l = (byte)(b[i] & 0xF);

                        bout[j] = (byte)ConvHexToAsc(h); //  (h >= 10 ? h + 0x41 - 10 : h + 0x30);
                        bout[j + 1] = (byte)ConvHexToAsc(l); // (l >= 10 ? l + 0x41 - 10 : l + 0x30);

                        j = j + 2;
                    }

                    // LRC の出力
                    lrc = (byte)(~lrc + 1);

                    h = (byte)((lrc & 0xF0) >> 4);
                    l = (byte)(lrc & 0xF);

                    bout[j] = (byte)ConvHexToAsc(h); // (h >= 10 ? h + 0x41 - 10 : h + 0x30);
                    bout[j + 1] = (byte)ConvHexToAsc(l);// (l >= 10 ? l + 0x41 - 10 : l + 0x30);
                    j = j + 2;

                    // CR LF の出力
                    bout[j] = 0xD;
                    bout[j + 1] = 0xA;

                    return bout;
                }
            }
            return null;
        }

        /// <summary>
        /// メンバー変数の初期化
        /// </summary>
        /// <remarks></remarks>
        private void clearData()
        {
            _dataFormat = null;
            _iLen = 0;
            _bStored = false;
            _eState = E_PARSE_STATE.EMPTY;
        }

        /// <summary>
    /// 初期化
    /// </summary>
    /// <remarks></remarks>
        public ClsDataPresentationAscii()
        {
            clearData();
        }

        /// <summary>
    /// バイト列による初期化
    /// </summary>
    /// <param name="b"></param>
    /// <remarks></remarks>
        public ClsDataPresentationAscii(ref byte[] b)
        {
            clearData();

            int i;
            var loopTo = b.Length - 1;
            for (i = 0; i <= loopTo; i++)
            {
                E_PARSE_STATE stat;

                stat = ParseByte(b[i]);

                if (stat == E_PARSE_STATE.COMPLETE)
                    break;
            }
        }

        public ClsDataPresentationAscii(ref ITweDataFormat fmt)
        {
            clearData();
            _dataFormat = fmt;
            _bStored = true;
            _eState = E_PARSE_STATE.COMPLETE;
        }


        public ITweDataFormat ObjFormat
        {
            get
            {
                return _dataFormat;
            }
        }

        public DateTime TimeGot
        {
            get
            {
                return _timeGot;
            }
            set
            {
                _timeGot = value;
            }
        }

        /// <summary>
    /// String型に変換するメソッド (System.Object.ToString() をオーバーライドする)
    /// ※ listBox などに格納したときに利用される。末尾にタイムスタンプを付加する。
    /// </summary>
    /// <returns></returns>
    /// <remarks></remarks>
        public override string ToString()
        {
            return System.Text.Encoding.ASCII.GetString(toByteArray()) + " (" + _timeGot.ToString() + ")";
        }

    }
}
